import requests


def urljoin(p1, p2):
    if not p1.endswith('/'):
        p1 += '/'
    return p1 + p2


class Keycloak:
    # TODO: add refresh_token
    PATH_TOKEN = "realms/{realm_name}/protocol/openid-connect/token"

    def __init__(self, base_url, client_id, realm_name):
        self.base_url = base_url
        self.client_id = client_id
        self.realm_name = realm_name
        self.session = requests.Session()

    def token(self, username, password):
        url = self.base_url + self.PATH_TOKEN.format(realm_name=self.realm_name)
        payload = {
            "username": username,
            "password": password,
            "client_id": self.client_id,
            "grant_type": ['password'],
        }
        return self.session.post(url, data=payload).json()

    def url(self, type):
        return urljoin(self.base_url, self.path(type))

    def path(self, type):
        return {
            'token': self.PATH_TOKEN.format(realm_name=self.realm_name),
        }[type]



import requests

from agilize.keycloak import Keycloak


class URL:
    BASE = 'https://app.agilize.com.br/api/v1/'

    DOWNLOAD_PROLABORE = BASE + 'companies/{company_id}/prolabore-anual/download'
    DOWNLOAD_TAX = BASE + 'companies/{company_id}/taxes/{tax_id}/billet'
    INFO = BASE + 'companies/security-user/info'
    INVOICES = BASE + 'companies/{company_id}/invoices'
    INVOICE_PAYMENT = BASE + 'companies/{company_id}/invoices/{invoice_id}'
    PARTNERS = BASE + 'companies/{company_id}/partners'
    PROLABORE = BASE + 'companies/{company_id}/prolabore-anual'
    TAXES = BASE + 'companies/{company_id}/taxes'
    UPLOAD_NFSE = BASE + 'companies/{company_id}/nfseimportresources'
    UPLOAD_NFSE2 = BASE + 'companies/{company_id}/nfses/importfromresource'


class AnonymousClient:
    @staticmethod
    def download(url):
        return requests.get(url).content


class Client(AnonymousClient):
    AUTH_URL = 'https://sso.agilize.com.br/auth/'
    CLIENT_ID = 'agilize-legacy-client'
    REALM_NAME = 'AgilizeAPPs'

    def __init__(self, username, password, keycloak=None):
        self.username = username
        self.password = password
        self.keycloak = keycloak or Keycloak(self.AUTH_URL, self.CLIENT_ID, self.REALM_NAME)
        self._access_token = None
        self._info = None

    @property
    def access_token(self):
        if not self._access_token:
            token = self.keycloak.token(self.username, self.password)
            self._access_token = token['access_token']
        return self._access_token

    @property
    def info(self):
        if not self._info:
            response = requests.get(
                url=URL.INFO,
                headers=self.headers,
            )
            self._info = response.json()
        return self._info

    @property
    def headers(self):
        return {'Authorization': f'Bearer {self.access_token}'}

    def partners(self, company_id):
        response = requests.get(
            url=URL.PARTNERS.format(company_id=company_id),
            headers=self.headers,
        )
        return response.json()

    def prolabores(self, company_id, year):
        response = requests.get(
            url=URL.PROLABORE.format(company_id=company_id),
            params={'anoReferencia': f'{year}-01-01T00:00:00P'},
            headers=self.headers,
        )
        return response.json()

    def download_prolabore(self, company_id, partner_id, year, month):
        response = requests.get(
            url=URL.DOWNLOAD_PROLABORE.format(company_id=company_id),
            params={
                'competence': f'{year}-{month}-01T00:00:00-0300',
                'partner': partner_id,
            },
            headers=self.headers,
        )
        return response.content

    def taxes(self, company_id, year):
        response = requests.get(
            url=URL.TAXES.format(company_id=company_id),
            params={
                'blocking': True,
                'closed': True,
                'count': 3000,
                'direction': 'desc',
                'onlyTaxesNotProvisionedByRh': True,
                'page': 1,
                'sort': 'companyTax.competence',
                'year': year,
            },
            headers=self.headers,
        )
        return response.json()

    def download_tax(self, company_id, tax_id):
        response = requests.get(
            url=URL.DOWNLOAD_TAX.format(company_id=company_id, tax_id=tax_id),
            headers=self.headers,
        )
        return self.download(response.json()['url'])

    def invoices(self, company_id, year):
        response = requests.get(
            url=URL.INVOICES.format(company_id=company_id),
            params={
                'count': 3000,
                'page': 1,
                'year': year,
            },
            headers=self.headers,
        )
        return response.json()

    def invoice_payment(self, company_id, invoice_id):
        response = requests.get(
            url=URL.INVOICE_PAYMENT.format(company_id=company_id, invoice_id=invoice_id),
            headers=self.headers,
        )
        return response.json()

    def upload_nfse(self, company_id, filebytes):
        response = requests.post(
            url=URL.UPLOAD_NFSE.format(company_id=company_id),
            files={'resources[0]': ('whatever.xml', filebytes, 'text/xml')},
            headers=self.headers,
        )
        response2 = requests.post(
            url=URL.UPLOAD_NFSE2.format(company_id=company_id),
            json={'nfseImportResource': response.json()['__identity']},
            headers=self.headers,
        )
        return response2.json()



import calendar
from collections import defaultdict
from datetime import date, datetime, timedelta
from decimal import Decimal
from typing import Optional

from attrs import define, field

from agilize.client import AnonymousClient, Client

DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%S%z'


class Agilize:
    def __init__(self, username, password, client=None):
        self.client = client or Client(username, password)

    def companies(self):
        return [
            Company.from_data(c, self.client)
            for c in self.client.info['party']['companies']
        ]


@define(kw_only=True)
class Prolabores:
    client: Client
    company_id: str
    _prolabores: dict = field(factory=dict)

    def get(self, competence):
        if competence not in self._prolabores:
            self.fetch(competence.year)

        return self._prolabores[competence]

    def fetch(self, year):
        # TODO: implement multiple partners
        data_by_date = self.client.prolabores(self.company_id, year)

        for datas in data_by_date.values():
            if not isinstance(datas, list):
                continue

            for d in datas:
                if not d['contraCheque']:
                    continue

                prolabore = Prolabore.from_data(d, self.company_id, self.client)
                self._prolabores[prolabore.competence] = prolabore

    def __iter__(self):
        yield from self._prolabores.values()


@define(kw_only=True)
class Taxes:
    client: Client
    company_id: str
    _taxes_by_competence: defaultdict = field(factory=lambda: defaultdict(dict))

    def get(self, abbreviation, competence):
        if competence not in self._taxes_by_competence:
            self.fetch(competence.year)

        return self._taxes_by_competence[competence][abbreviation]

    def fetch(self, year):
        data = self.client.taxes(self.company_id, year)

        for d in data:
            tax = Tax.from_data(d, self.company_id, self.client)
            self._taxes_by_competence[tax.competence][tax.abbreviation] = tax

    def filter(self, competence):
        if competence not in self._taxes_by_competence:
            self.fetch(competence.year)

        return list(self._taxes_by_competence[competence].values())

    def __iter__(self):
        for t in self._taxes_by_competence.values():
            yield from t.values()


@define(kw_only=True)
class Invoices:
    client: Client
    company_id: str
    _invoices: dict = field(factory=dict)

    def get(self, competence):
        if competence not in self._invoices:
            self.fetch(competence.year)

        return self._invoices[competence]

    def fetch(self, year):
        data = self.client.invoices(self.company_id, year)

        for d in data:
            if not d['nfses']:
                continue
            invoice = Invoice.from_data(d, self.company_id, self.client)
            self._invoices[invoice.competence] = invoice

    def __iter__(self):
        yield from self._invoices.values()


@define
class Company:
    id: str
    cnpj: str
    name: str
    client: Client
    _invoices: Optional[Invoices] = None
    _prolabores: Optional[Prolabores] = None
    _taxes: Optional[Taxes] = None

    @classmethod
    def from_data(cls, data, client):
        return cls(
            id=data['__identity'],
            cnpj=data['cnpj'],
            name=data['name'],
            client=client,
        )

    @property
    def invoices(self):
        if not self._invoices:
            self._invoices = Invoices(client=self.client, company_id=self.id)
        return self._invoices

    @property
    def prolabores(self):
        if not self._prolabores:
            self._prolabores = Prolabores(client=self.client, company_id=self.id)
        return self._prolabores

    @property
    def taxes(self):
        if not self._taxes:
            self._taxes = Taxes(client=self.client, company_id=self.id)
        return self._taxes

    def upload_nfse(self, filebytes):
        data = self.client.upload_nfse(self.id, filebytes)
        return data['countNfses']


@define(hash=True)
class Competence:
    year: int = field(converter=int)
    month: int = field(converter=int)

    @classmethod
    def from_data(cls, data):
        year, month, *_ = data.split('-')
        return cls(int(year), int(month))

    @classmethod
    def from_date(cls, date):
        return cls(date.year, date.month)

    @property
    def first_date(self):
        return date(self.year, self.month, 1)

    @property
    def last_date(self):
        _, last_day = calendar.monthrange(self.year, self.month)
        return date(self.year, self.month, last_day)

    @property
    def next(self):
        return self.from_date(self.last_date + timedelta(days=1))

    @property
    def previous(self):
        return self.from_date(self.first_date - timedelta(days=1))

    def __str__(self):
        return f'{self.year:04d}{self.month:02d}'


@define
class Prolabore:
    client: Client
    company_id: str
    competence: Competence
    inss: Decimal
    irpf: Decimal
    net_value: Decimal
    partner_id: str
    paycheck_id: str
    total_value: Decimal

    @classmethod
    def from_data(cls, data, company_id, client):
        return cls(
            client=client,
            company_id=company_id,
            competence=Competence.from_data(data['competence']),
            inss=Decimal(str(data['iNSS'])),
            irpf=Decimal(str(data['iRPJFolha'])),
            net_value=Decimal(str(data['valorLiquido'])),
            partner_id=data['partner']['__identity'],
            paycheck_id=data['contraCheque']['__identity'],
            total_value=Decimal(str(data['valor'])),
        )

    def download(self):
        return self.client.download_prolabore(
            self.company_id,
            self.partner_id,
            self.competence.year,
            self.competence.month,
        )


@define
class Tax:
    abbreviation: str
    amount: Decimal
    client: Client
    company_id: str
    competence: Competence
    due_date: date
    id: str

    @classmethod
    def from_data(cls, data, company_id, client):
        return cls(
            abbreviation=data['taxAbbreviation'],
            amount=Decimal(str(data['total'])),
            client=client,
            company_id=company_id,
            competence=Competence.from_data(data['competence']),
            due_date=datetime.strptime(data['deadline'], DATETIME_FORMAT).date(),
            id=data['__identity'],
        )

    def download(self):
        return self.client.download_tax(self.company_id, self.id)


@define
class Invoice:
    amount: Decimal
    company_id: str
    competence: Competence
    due_date: date
    id: str
    url_nfse: str
    client: Client
    _barcode: Optional[str] = None

    @classmethod
    def from_data(cls, data, company_id, client):
        return cls(
            amount=Decimal(str(data['total'])),
            company_id=company_id,
            competence=Competence.from_data(data['competence']),
            client=client,
            due_date=datetime.strptime(data['deadline'], DATETIME_FORMAT).date(),
            id=data['__identity'],
            url_nfse=data['nfses'][0]['nfseUrl'],
        )

    @property
    def barcode(self):
        if not self._barcode:
            data = self.client.invoice_payment(self.company_id, self.id)
            self._barcode = data['validBilletPaymentOrder']['billet']['barcode']
        return self._barcode

    @property
    def url_nfse_image(self):
        return self.url_nfse.replace('notaprint.aspx', 'notaprintimg.aspx')

    def download_nfse(self):
        return AnonymousClient.download(self.url_nfse_image)



import os

from setuptools import setup

README = os.path.join(os.path.dirname(__file__), 'README.md')


if __name__ == "__main__":
    setup(
        name='Agilize',
        description='Unofficial client to access Agilize',
        version='0.0.10',
        long_description=open(README).read(),
        long_description_content_type='text/markdown',
        author="Lucas Rangel Cezimbra",
        author_email="lucas@cezimbra.tec.br",
        license="LGPLv2",
        url='https://github.com/lucasrcezimbra/agilize',
        keywords=['agilize', 'api', 'client', 'requests', 'accounting', 'finance'],
        install_requires=['attrs', 'requests'],
        packages=['agilize'],
        zip_safe=False,
        include_package_data=True,
        classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3 :: Only',
            'Programming Language :: Python :: Implementation :: CPython',
            'Topic :: Office/Business :: Financial',
            'Topic :: Office/Business :: Financial :: Accounting',
        ],
    )
